import gymnasium
from gymnasium import spaces
import numpy as np
import time
from PIL import ImageChops
from eventstreaming import stream

class DinoEnv(gymnasium.Env):
    metadata = { 'render_modes': [], 'render_fps': 4 }

    def __init__(self, render_mode=None):
        self.observation_space = spaces.Dict(
            {
                'screen': spaces.Box(low=0, high=255, shape=(36, 144, 3), dtype=np.uint8)
            }
        )

        # We have 3 actions, corresponding to 0: "none", 1: "up", 2: "down"
        # self.action_space = spaces.Discrete(3)
        self.action_space = spaces.Discrete(2)

        assert render_mode is None or render_mode in self.metadata['render_modes']
        self.render_mode = render_mode

        self.prev_img = None
        self.img = None
        self.count = 0

        self.pause = False

        stream.start(top=330, width=1440, height=360)

    def reset(self, seed=None, options=None):
        inputs = []
        while self.pause:
            time.sleep(0.01)
            inputs = stream.get_io_events(timestamp=False)
            if 'p esc' in inputs:
                self.pause = not self.pause

        # We need the following line to seed self.np_random
        super().reset(seed=seed)

        # Focus on the game
        stream.send_io_event('m200,860')
        stream.send_io_event('cpleft')
        stream.send_io_event('crleft')
        time.sleep(1)
        stream.send_io_event('p space')
        stream.send_io_event('r space')

        frame = stream.get_video_frame()
        while frame is None:
            frame = stream.get_video_frame()
        self.img = frame['image']
        img = self.img
        self.prev_img = img
        
        # observation = { 'screen': np.array(img) }
        observation = { 'screen': np.array(img.resize((144, 36))) }

        return observation, {}

    def step(self, action):
        inputs = stream.get_io_events(timestamp=False)
        if 'p esc' in inputs:
            self.pause = not self.pause
        
        if action == 0 or self.count > 0:
            pass
        elif action == 1:
            stream.send_io_event('p up')
            time.sleep(0.2)
            stream.send_io_event('r up')
        # elif action == 2:
        #     stream.send_io_event('p down')
        #     time.sleep(0.5)
        #     stream.send_io_event('r down')

        time.sleep(0.1)
        frame = stream.get_video_frame()
        if frame is not None:
            self.img = frame['image']
        img = self.img
        
        reward = 0
        
        terminated = False
        diff = ImageChops.difference(self.img, self.prev_img)
        if not diff.getbbox():
            self.count += 1
        if self.count > 5:
            self.count = 0
            terminated = True
        self.prev_img = self.img
        
        # observation = { 'screen': np.array(img) }
        observation = { 'screen': np.array(img.resize((144, 36))) }

        return observation, reward, terminated, self.pause, {}

    def render(self):
        pass

    def close(self):
        stream.close()


import gymnasium
from gymnasium import spaces
import pytesseract
import numpy as np
import time
from eventstreaming.websocket import stream

class DinoEnv(gymnasium.Env):
    metadata = { "render_modes": [], "render_fps": 4 }

    def __init__(self, render_mode=None):
        self.observation_space = spaces.Box(0, 255, shape=(36, 144, 1), dtype=np.uint8)

        # We have 3 actions, corresponding to 0: "none", 1: "up", 2: "down"
        # self.action_space = spaces.Discrete(3)
        self.action_space = spaces.Discrete(2)

        assert render_mode is None or render_mode in self.metadata["render_modes"]
        self.render_mode = render_mode

        self.prev_score = 0
        self.img = None

        self.pause = False

        stream.start()

    def reset(self, seed=None, options=None):
        inputs = []
        while self.pause:
            time.sleep(0.01)
            inputs = stream.get_inputs()
            if 'p esc' in inputs:
                self.pause = not self.pause

        # We need the following line to seed self.np_random
        super().reset(seed=seed)

        self.prev_score = 0

        # Focus on the game
        stream.send_io_event('m200,860')
        stream.send_io_event('cpleft')
        stream.send_io_event('crleft')
        time.sleep(1)
        stream.send_io_event('p space')
        stream.send_io_event('r space')

        frame = stream.get_video_frame()
        while frame is None:
            frame = stream.get_video_frame()
        self.img = frame['image']
        img = self.img
        
        img = img.resize((144, 36)).convert('L')
        observation = np.array(img)[..., np.newaxis]

        return observation, {}

    def step(self, action):
        inputs = stream.get_io_events(timestamp=False)
        if 'p esc' in inputs:
            self.pause = not self.pause
        
        if action == 0:
            pass
        elif action == 1:
            stream.send_io_event('p up')
            # time.sleep(0.2)
            stream.send_io_event('r up')
        # elif action == 2:
        #     stream.send_io_event('p down')
        #     time.sleep(0.5)
        #     stream.send_io_event('r down')

        frame = stream.get_video_frame()
        if frame is not None:
            self.img = frame['image']
        img = self.img
        
        # texts = pytesseract.image_to_string(img.crop((1100, 0, 1440, 60))).strip().split(" ")
        # score = 0
        # if len(texts) == 2:
        #     try:
        #         highest = int(texts[0])
        #         score = int(texts[1])
        #     except ValueError:
        #         score = self.prev_score + 1000
        #     if score - self.prev_score > 100:
        #         score = self.prev_score + 1
        # reward = score - self.prev_score
        # if reward < 0:
        #     reward = 0
        #     score = self.prev_score
        # self.prev_score = score
        reward = 0
        
        text = pytesseract.image_to_string(img.crop((480, 60, 960, 180))).strip()
        terminated = False
        if 'GAME' in text:
            terminated = True
        
        img = img.resize((144, 36)).convert('L')
        observation = np.array(img)[..., np.newaxis]

        return observation, reward, terminated, self.pause, {}

    def render(self):
        pass

    def close(self):
        stream.close()


import gymnasium
from gymnasium import spaces
import numpy as np
import time
from eventstreaming import stream
from screen_games.config import bool_ENV, int_ENV

class ScreenEnv(gymnasium.Env):
    metadata = { 'render_modes': [], 'render_fps': 30 }

    def __init__(self, render_mode=None, left=None, top=None, width=None, height=None, new_width=None, new_height=None, fullscreen=None):
        if left is None:
            left = int_ENV('SCREENENV_LEFT', 0)
        if top is None:
            top = int_ENV('SCREENENV_TOP', 0)
        if width is None:
            width = int_ENV('SCREENENV_WIDTH', -1)
            if width == -1: width = None
        if height is None:
            height = int_ENV('SCREENENV_HEIGHT', -1)
            if height == -1: height = None
        if new_width is None:
            new_width = int_ENV('SCREENENV_NEW_WIDTH', -1)
            if new_width == -1: new_width = width
        if new_height is None:
            new_height = int_ENV('SCREENENV_NEW_HEIGHT', -1)
            if new_height == -1: new_height = height
        if fullscreen is None:
            fullscreen = bool_ENV('SCREENENV_FULLSCREEN', False)
        
        self.observation_space = spaces.Dict(
            {
                'screen': spaces.Box(low=0, high=255, shape=(new_height, new_width, 3), dtype=np.uint8),
                'timestamp': spaces.Box(low=0, high=np.inf, shape=(1,), dtype=np.float32)
            }
        )

        self.action_space = spaces.Discrete(2)

        assert render_mode is None or render_mode in self.metadata['render_modes']
        self.render_mode = render_mode

        self.frame = None
        self.start_time = 0

        self.pause = False

        stream.start(left=left, top=top, width=width, height=height, new_width=new_width, new_height=new_height, fullscreen=fullscreen)

    def reset(self, seed=None, options=None, on_reset=None):
        inputs = []
        while self.pause:
            time.sleep(1.0 / ScreenEnv.metadata['render_fps'])
            inputs = stream.get_io_events()
            if 'p esc' in list(map(lambda x: x['event'], inputs)):
                self.pause = not self.pause
        
        # We need the following line to seed self.np_random
        super().reset(seed=seed)

        if on_reset is not None:
            on_reset()

        frame = stream.get_video_frame()
        while frame is None:
            frame = stream.get_video_frame()
        self.frame = frame
        self.start_time = self.frame['timestamp'] / 1_000_000.0
        
        observation = {
            'screen': np.array(self.frame['image']),
            'timestamp': np.array([self.start_time], dtype=np.float32)
        }

        return observation, { 'inputs': inputs }

    def step(self, action):
        inputs = stream.get_io_events()
        if 'p esc' in list(map(lambda x: x['event'], inputs)):
            self.pause = not self.pause

        time.sleep(1.0 / ScreenEnv.metadata['render_fps'])
        frame = stream.get_video_frame()
        if frame is not None:
            self.frame = frame
        
        observation = {
            'screen': np.array(self.frame['image']),
            'timestamp': np.array([self.frame['timestamp'] / 1_000_000.0 - self.start_time], dtype=np.float32)
        }

        return observation, 0, False, self.pause, { 'inputs': inputs }

    def render(self):
        pass

    def close(self):
        stream.close()